import sqlite3

import pytest
from pyogrio.errors import FieldError

from geost import utils


@pytest.mark.unittest
def test_save_pickle(borehole_collection, tmp_path):
    outfile = tmp_path / r"collection.pkl"
    utils.save_pickle(borehole_collection, outfile)
    assert outfile.is_file()


@pytest.mark.unittest
def test_to_geopackage(borehole_collection, tmp_path):
    outfile = tmp_path / "gpkg_with_error.gpkg"
    invalid_column = "GEOM"
    borehole_collection.header[invalid_column] = "test"
    with pytest.raises(FieldError):
        borehole_collection.to_geopackage(outfile)


@pytest.mark.unittest
def test_create_connection(simple_soilmap_gpkg):
    conn = utils.create_connection(simple_soilmap_gpkg)
    assert isinstance(conn, sqlite3.Connection)



import pytest
from lxml.etree import _Element

from geost.bro import BroApi


class TestBroApi:
    @pytest.mark.unittest
    def test_response(self):
        api = BroApi()
        for key in api.apis:
            # Exclude BRO Geological boreholes until added to BRO database
            # (They currently return a 503 status code)
            if key != "BHR-G":
                assert (
                    api.session.get(api.server_url + api.apis[key]).status_code == 200
                )

    @pytest.mark.unittest
    def test_get_valid_cpts(self):
        api = BroApi()
        cpt_datas = api.get_objects(
            ["CPT000000038771", "CPT000000000787", "CPT000000125133"], object_type="CPT"
        )
        for cpt_data in cpt_datas:
            assert isinstance(cpt_data, _Element)

    @pytest.mark.unittest
    def test_get_valid_bhrps(self):
        api = BroApi()
        bhr_datas = api.get_objects(
            ["BHR000000342629", "BHR000000195688", "BHR000000206176"],
            object_type="BHR-P",
        )
        for bhr_data in bhr_datas:
            assert isinstance(bhr_data, _Element)

    @pytest.mark.unittest
    def test_get_valid_bhrgts(self):
        api = BroApi()
        bhr_datas = api.get_objects(
            ["BHR000000347437", "BHR000000347438", "BHR000000347434"],
            object_type="BHR-GT",
        )
        for bhr_data in bhr_datas:
            assert isinstance(bhr_data, _Element)

    @pytest.mark.unittest
    def test_get_valid_bhrgs(self):
        pass
        # BHR-G objects are not yet available in the BRO

        # api = BroApi()
        # bhr_datas = api.get_objects(
        #     ["BHR000000342629", "BHR000000195688", "BHR000000206176"],
        #     object_type="BHR-G",
        # )
        # for bhr_data in bhr_datas:
        #     assert isinstance(bhr_data, _Element)

    @pytest.mark.unittest
    def test_get_invalid_cpt(self):
        api = BroApi()
        cpt_datas = api.get_objects("CPT0000000doesnotexist", object_type="CPT")
        with pytest.raises(Warning) as excinfo:
            for cpt_data in cpt_datas:
                pass
        assert "Unable to request CPT0000000doesnotexist" in str(excinfo.value)

    @pytest.mark.unittest
    def test_get_invalid_bhr(self):
        api = BroApi()
        bhr_datas = api.get_objects("BHR0000000doesnotexist", object_type="BHR-GT")
        with pytest.raises(Warning) as excinfo:
            for bhr_data in bhr_datas:
                pass
        assert "Unable to request BHR0000000doesnotexist" in str(excinfo.value)

    @pytest.mark.unittest
    def test_search_and_get_cpts_in_bbox(self):
        # Note: if test fails, check if BRO database was updated first
        api = BroApi()
        # Find CPTs
        api.search_objects_in_bbox(112400, 442750, 112500, 442850, object_type="CPT")
        minimum_present_objects = ["CPT000000000787", "CPT000000029403"]
        present_objects = [
            obj for obj in api.object_list if obj in minimum_present_objects
        ]
        assert len(present_objects) == len(minimum_present_objects)
        # Get CPTs
        cpt_datas = api.get_objects(api.object_list, object_type="CPT")
        for cpt_data in cpt_datas:
            assert isinstance(cpt_data, _Element)

    @pytest.mark.unittest
    def test_search_and_get_bhrps_in_bbox(self):
        # Note: if test fails, check if BRO database was updated first
        api = BroApi()
        # Find CPTs
        api.search_objects_in_bbox(141500, 455100, 141700, 455300, object_type="BHR-P")
        minimum_present_objects = [
            "BHR000000085497",
            "BHR000000247842",
            "BHR000000120513",
            "BHR000000206176",
        ]
        present_objects = [
            obj for obj in api.object_list if obj in minimum_present_objects
        ]
        assert len(present_objects) == len(minimum_present_objects)
        # Get CPTs
        bhr_datas = api.get_objects(api.object_list, object_type="BHR-P")
        for bhr_data in bhr_datas:
            assert isinstance(bhr_data, _Element)

    @pytest.mark.unittest
    def test_search_and_get_bhrgts_in_bbox(self):
        # Note: if test fails, check if BRO database was updated first
        api = BroApi()
        # Find CPTs
        api.search_objects_in_bbox(141300, 452700, 142300, 453500, object_type="BHR-GT")
        minimum_present_objects = [
            "BHR000000353592",
            "BHR000000353583",
            "BHR000000353598",
            "BHR000000353600",
        ]
        present_objects = [
            obj for obj in api.object_list if obj in minimum_present_objects
        ]
        assert len(present_objects) == len(minimum_present_objects)
        # Get CPTs
        bhr_datas = api.get_objects(api.object_list, object_type="BHR-GT")
        for bhr_data in bhr_datas:
            assert isinstance(bhr_data, _Element)

    @pytest.mark.unittest
    def test_get_too_large_volume(self, capsys):
        api = BroApi()
        cpts = api.search_objects_in_bbox(82000, 425000, 90000, 432000)
        captured = capsys.readouterr()
        assert "More than 2000 object requests in API call" in captured.out



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from geost import (
    get_bro_objects_from_bbox,
    get_bro_objects_from_geometry,
    read_borehole_table,
    read_collection_geopackage,
    read_cpt_table,
    read_gef_cpts,
    read_nlog_cores,
    read_pickle,
    read_uullg_tables,
    read_xml_boris,
)
from geost.base import (
    BoreholeCollection,
    CptCollection,
    DiscreteData,
    LayeredData,
    PointHeader,
)
from geost.read import (
    MANDATORY_LAYERED_DATA_COLUMNS,
    _check_mandatory_column_presence,
    adjust_z_coordinates,
)


class InvalidCollection:
    """
    Invalid Collection type to raise a ValueError in `read_collection_geopackage`.
    """

    pass


@pytest.fixture
def data_dir():
    return Path(__file__).parent / "data"


@pytest.fixture
def llg_header_table(tmp_path):
    outfile = tmp_path / r"llg_header.parquet"
    header = pd.DataFrame(
        {
            "BOORP": [1901, 1902],
            "XCO": [1, 2],
            "YCO": [1, 2],
            "MV_HoogteNAP": [2.1, 1.9],
            "BoringEinddiepteNAP": [1.7, 1.5],
        }
    )
    header.to_parquet(outfile)
    return outfile


@pytest.fixture
def llg_data_table(tmp_path):
    outfile = tmp_path / r"llg_data.parquet"
    data = pd.DataFrame(
        {
            "BOORP": [1901, 1901, 1902, 1902],
            "BEGIN_DIEPTE": [0, 30, 0, 20],
            "EIND_DIEPTE": [30, 40, 20, 40],
            "TEXTUUR": ["ZK", "L", "Z", "K"],
        }
    )
    data.to_parquet(outfile)
    return outfile


@pytest.fixture
def llg_data_table_duplicate_column(tmp_path):
    outfile = tmp_path / r"llg_data_dupl.parquet"
    data = pd.DataFrame(
        {
            "nr": [1901, 1901, 1902, 1902],
            "top": [0, 30, 0, 20],
            "bottom": [30, 40, 20, 40],
            "TEXTUUR": ["ZK", "L", "Z", "K"],
        }
    )
    data.to_parquet(outfile)
    return outfile


@pytest.fixture
def collection_pickle(borehole_collection, tmp_path):
    outfile = tmp_path / r"collection.pkl"
    borehole_collection.to_pickle(outfile)
    return outfile


class TestReaders:
    @pytest.fixture
    def table_wrong_columns(self):
        return pd.DataFrame(
            {
                "nr": ["a", "b"],
                "x": [1, 2],
                "y": [1, 2],
                "maaiveld": [1, 2],
                "end": [1, 1],
                "top": [1, 1],
                "bottom": [2, 2],
            }
        )

    @pytest.mark.unittest
    def test_nlog_reader_from_parquet(self, data_dir):
        nlog_cores = read_nlog_cores(
            data_dir / r"test_nlog_stratstelsel_20230807.parquet"
        )
        desired_df = pd.DataFrame(
            {
                "nr": ["BA050018", "BA080036", "BA110040"],
                "x": [37395, 44175, 39309],
                "y": [857077, 840614, 833198],
                "surface": [36.7, 39.54, 33.51],
                "end": [-3921.75, -3262.69, -3865.89],
            }
        )
        assert_array_equal(
            nlog_cores.header[["nr", "x", "y", "surface", "end"]], desired_df
        )
        assert nlog_cores.data.has_inclined

    @pytest.mark.unittest
    def test_get_bro_soil_cores_from_bbox(self):
        soilcores = get_bro_objects_from_bbox(
            "BHR-P", xmin=87000, ymin=444000, xmax=87500, ymax=444500
        )
        assert soilcores.n_points == 7

    @pytest.mark.unittest
    def test_get_bro_soil_cores_from_geometry(self, data_dir):
        soilcores = get_bro_objects_from_geometry(
            "BHR-P", data_dir / "test_polygon.parquet"
        )
        assert soilcores.n_points == 1

    @pytest.mark.unittest
    def test_read_borehole_table(self, data_dir):
        file_pq = data_dir / r"test_borehole_table.parquet"
        file_csv = data_dir / r"test_borehole_table.csv"
        cores_pq = read_borehole_table(file_pq)
        cores_csv = read_borehole_table(file_csv)
        assert isinstance(cores_pq, BoreholeCollection)
        assert isinstance(cores_csv, BoreholeCollection)

        cores_pq = read_borehole_table(file_pq, as_collection=False)
        cores_csv = read_borehole_table(file_csv, as_collection=False)
        assert isinstance(cores_pq, LayeredData)
        assert isinstance(cores_csv, LayeredData)

    @pytest.mark.unittest
    def test_read_inclined_borehole_table(self, data_dir):
        file_pq = data_dir / r"test_inclined_borehole_table.parquet"
        cores_pq = read_borehole_table(file_pq, has_inclined=True)
        assert isinstance(cores_pq, BoreholeCollection)

        cores_pq = read_borehole_table(file_pq, has_inclined=True, as_collection=False)
        assert isinstance(cores_pq, LayeredData)

    @pytest.mark.unittest
    def test_check_mandatory_columns(self, table_wrong_columns):
        column_mapper = {"maaiveld": "surface"}
        table_wrong_columns = _check_mandatory_column_presence(
            table_wrong_columns, MANDATORY_LAYERED_DATA_COLUMNS, column_mapper
        )
        assert_array_equal(table_wrong_columns.columns, MANDATORY_LAYERED_DATA_COLUMNS)

    @pytest.mark.unittest
    def test_check_mandatory_columns_with_user_input(
        self, table_wrong_columns, monkeypatch
    ):
        monkeypatch.setattr("builtins.input", lambda _: "maaiveld")
        table_wrong_columns = _check_mandatory_column_presence(
            table_wrong_columns, MANDATORY_LAYERED_DATA_COLUMNS
        )
        assert_array_equal(table_wrong_columns.columns, MANDATORY_LAYERED_DATA_COLUMNS)

    @pytest.mark.unittest
    def test_adjust_z_coordinates(self, data_dir):
        file = data_dir / r"test_borehole_table.parquet"
        cores = read_borehole_table(file, as_collection=False)
        cores_df = cores.df.copy()

        # test situation where top and bottoms are already positive downward
        cores_df_adjusted = adjust_z_coordinates(cores_df.copy())
        assert cores_df_adjusted["top"].iloc[0] < cores_df_adjusted["top"].iloc[1]
        assert cores_df_adjusted["bottom"].iloc[0] < cores_df_adjusted["bottom"].iloc[1]

        # Test situation where top and bottoms are given as negative downward
        cores_df["top"] *= -1
        cores_df["bottom"] *= -1

        cores_df_adjusted = adjust_z_coordinates(cores_df.copy())
        assert cores_df_adjusted["top"].iloc[0] < cores_df_adjusted["top"].iloc[1]
        assert cores_df_adjusted["bottom"].iloc[0] < cores_df_adjusted["bottom"].iloc[1]

    @pytest.mark.unittest
    def test_read_boris_xml(self, data_dir):
        boris_collection = read_xml_boris(data_dir / r"test_boris_xml.xml")
        assert isinstance(boris_collection, BoreholeCollection)
        assert boris_collection.n_points == 16
        assert len(boris_collection.data.df) == 236


@pytest.mark.unittest
def test_read_uullg_table(
    llg_header_table, llg_data_table, llg_data_table_duplicate_column
):
    llg = read_uullg_tables(llg_header_table, llg_data_table)
    assert isinstance(llg, BoreholeCollection)
    expected_header_cols = ["nr", "x", "y", "surface", "end"]
    header_columns = llg.header.gdf.columns
    assert all([c in header_columns for c in expected_header_cols])
    expected_data_cols = ["nr", "x", "y", "surface", "end"]
    data_columns = llg.data.df.columns
    assert all([c in data_columns for c in expected_data_cols])

    llg = read_uullg_tables(llg_header_table, llg_data_table_duplicate_column)
    assert isinstance(llg, BoreholeCollection)
    expected_header_cols = ["nr", "x", "y", "surface", "end"]
    header_columns = llg.header.gdf.columns
    assert all([c in header_columns for c in expected_header_cols])
    expected_data_cols = ["nr", "x", "y", "surface", "end"]
    data_columns = llg.data.df.columns
    assert all([c in data_columns for c in expected_data_cols])


@pytest.mark.unittest
def test_read_gef_cpts(data_dir):
    files = sorted(Path(data_dir / "cpt").glob("*.gef"))
    cpts = read_gef_cpts(files)
    assert isinstance(cpts, CptCollection)

    expected_cpts_present = [
        "DKMP_D03",
        "AZZ158",
        "CPT000000157983",
        "YANGTZEHAVEN CPT 10",
    ]
    assert_array_equal(cpts.header["nr"], expected_cpts_present)


@pytest.mark.unittest
def test_read_cpt_table(data_dir, monkeypatch):
    monkeypatch.setattr("builtins.input", lambda _: "mv")
    cpts = read_cpt_table(data_dir / r"test_cpts.parquet")
    assert isinstance(cpts, CptCollection)
    assert cpts.horizontal_reference == 28992
    assert cpts.vertical_reference == 5709

    cpts = read_cpt_table(data_dir / r"test_cpts.parquet", as_collection=False)
    assert isinstance(cpts, DiscreteData)


@pytest.mark.unittest
def test_read_pickle(collection_pickle):
    collection = read_pickle(collection_pickle)
    assert isinstance(collection, BoreholeCollection)


class TestReadCollectionGeopackage:
    @pytest.fixture
    def borehole_gpkg(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"borehole.gpkg"
        borehole_collection.to_geopackage(outfile)
        return outfile

    @pytest.fixture
    def cpt_gpkg(self, cpt_collection, tmp_path):
        outfile = tmp_path / r"cpt.gpkg"
        cpt_collection.to_geopackage(outfile)
        return outfile

    @pytest.mark.unittest
    def test_read_collection_geopackage_boreholes(self, borehole_gpkg):
        collection = read_collection_geopackage(borehole_gpkg, BoreholeCollection)
        assert isinstance(collection, BoreholeCollection)
        assert isinstance(collection.header, PointHeader)
        assert isinstance(collection.data, LayeredData)

    @pytest.mark.unittest
    def test_read_collection_geopackage_cpts(self, cpt_gpkg):
        collection = read_collection_geopackage(cpt_gpkg, CptCollection)
        assert isinstance(collection, CptCollection)
        assert isinstance(collection.header, PointHeader)
        assert isinstance(collection.data, DiscreteData)

    @pytest.mark.unittest
    def test_read_collection_geopackage_invalid(self, borehole_gpkg):
        with pytest.raises(ValueError):
            read_collection_geopackage(borehole_gpkg, InvalidCollection)



from functools import wraps

from geost.validate.validate import DataFrameSchema
from geost.validate.validation_schemes import ValidationSchemas

validationschemas = ValidationSchemas()


def validate_header(f):
    @wraps(f)
    def wrapper(*args):
        dataframe_to_validate = args[1]
        validation_instance = DataFrameSchema(
            "Header", validationschemas.headerschema_point
        )
        validation_instance.validate(dataframe_to_validate)
        return f(*args)

    return wrapper


def validate_data(f):
    @wraps(f)
    def wrapper(*args):
        data_object = args[0]
        dataframe_to_validate = args[1]
        dtype = data_object.datatype

        # Determine what the validation conditions are and append to 'schema_to_use'
        if dtype == "layered":
            schema_to_use = validationschemas.dataschema_layered_point
        elif dtype == "discrete":
            schema_to_use = validationschemas.dataschema_discrete_point

        if data_object.has_inclined:
            schema_to_use = schema_to_use | validationschemas.dataschema_inclined_point

        # Use the acquired validation conditions for validation
        validation_instance = DataFrameSchema("Data", schema_to_use)
        validation_instance.validate(dataframe_to_validate)

        return f(*args)

    return wrapper



from typing import Iterable

import numpy as np
import pandas as pd
import pyvista as pv
import xarray as xr
from pyvista import CellType, MultiBlock


def prepare_borehole(borehole: pd.DataFrame, vertical_factor: float) -> np.ndarray:
    borehole_xyz = borehole[["x", "y", "bottom"]].to_numpy()
    surface_xyz = np.array(
        [borehole_xyz[0, 0], borehole_xyz[0, 1], borehole["surface"].iloc[0]]
    )
    borehole_xyz = np.vstack([surface_xyz, borehole_xyz], dtype="float64")
    borehole_xyz[:, 2] *= vertical_factor
    return borehole_xyz


def generate_cylinders(
    table: pd.DataFrame,
    data_columns: list[str],
    radius: float,
    vertical_factor: float,
) -> Iterable:
    boreholes = table.groupby("nr")
    for _, borehole in boreholes:
        borehole_prepared = prepare_borehole(borehole, vertical_factor)
        poly = pv.PolyData(borehole_prepared)
        line_segments = np.arange(0, len(borehole_prepared), dtype=np.int_)
        line_segments = np.insert(line_segments, 0, len(borehole_prepared))
        poly.lines = line_segments

        for data_column in data_columns:
            poly[data_column] = np.hstack(
                [borehole[data_column].values[0], borehole[data_column].values]
            )
        cylinder = poly.tube(radius=radius)
        yield cylinder


def borehole_to_multiblock(
    table: pd.DataFrame,
    data_columns: list[str],
    radius: float,
    vertical_factor: float,
) -> MultiBlock:
    """
    Create a PyVista MultiBlock object from the parsed boreholes/cpt's.

    Parameters
    ----------
    table : pd.DataFrame
        Table of borehole/CPT objects. This is CptCollection.data or
        BoreholeCollection.data.
    data_columns : List[str]
        Column names of data arrays to write in the vtk file
    radius : float
        Radius of borehole cylinders
    vertical_factor : float
        Vertical adjustment factor to convert e.g. heights in cm to m.

    Returns
    -------
    pv.MultiBlock
        MultiBlock object with boreholes represented as cylinder geometries

    """
    cylinders = generate_cylinders(table, data_columns, radius, vertical_factor)
    cylinders_multiblock = pv.MultiBlock(list(cylinders))
    return cylinders_multiblock


def voxelmodel_to_pyvista_unstructured(
    dataset: xr.Dataset,
    resolution: tuple[float, float, float],
    displayed_variables: list[str] = None,
) -> pv.UnstructuredGrid:
    """
    Convert an xarray dataset to a PyVista voxel model (UnstructuredGrid).

    Parameters
    ----------
    dataset : xarray.Dataset
        The input dataset containing 3D grid data with coordinates 'x', 'y', and 'z'.
    displayed_variables : list of str
        List of variable names in the dataset to include as cell data in the voxel model.
    resolution : tuple of float
        The resolution (dx, dy, dz) representing the size of each voxel along the x, y,
        and z axes.

    Returns
    -------
    grid : pyvista.UnstructuredGrid
        The resulting voxel model as a PyVista UnstructuredGrid, with cell data assigned
        for each variable in `displayed_variables`.
    """
    if displayed_variables is None:
        displayed_variables = dataset.data_vars

    if "z" not in dataset.coords:
        raise ValueError(
            "Dataset must contain 'z' dimension. Use xarray.Dataset.rename() to rename "
            + "the corresponding dimension to 'z' if applicable."
        )

    # Order dataset dimensions to match required order
    expected_order = ("y", "x", "z")
    if tuple(dataset.dims.keys()) != expected_order:
        dataset = dataset.transpose(*expected_order)

    # Get half resolution (used to compute voxel corners with respect to cell centers)
    xres, yres, zres = resolution[0] / 2, resolution[1] / 2, resolution[2] / 2

    # Compute cell centers (Making use of the fact that xarray coordinates represent
    # those cell centers, so we can use them directly with meshgrid)
    xx, yy, zz = np.meshgrid(dataset.x.values, dataset.y.values, dataset.z.values)
    cell_centers = np.stack([xx, yy, zz], axis=-1).reshape(-1, 3)

    # Compute voxel corners by offsetting cell centers with half resolution in all
    # directions.
    offsets = np.array(
        [
            [-xres, -yres, -zres],
            [xres, -yres, -zres],
            [-xres, yres, -zres],
            [xres, yres, -zres],
            [-xres, -yres, zres],
            [xres, -yres, zres],
            [-xres, yres, zres],
            [xres, yres, zres],
        ]
    )
    voxels = cell_centers[:, None, :] + offsets[None, :, :]
    points = voxels.reshape(-1, 3)

    # Build cell connectivity
    n_voxels = cell_centers.shape[0]
    cells_voxel = np.arange(n_voxels * 8).reshape((n_voxels, 8))

    # Create unstructured grid and assign data variables
    grid = pv.UnstructuredGrid({CellType.VOXEL: cells_voxel}, points)
    for var in displayed_variables:
        if not all(dim in dataset[var].dims for dim in ["x", "y", "z"]):
            print(
                f"Variable '{var}' does not have the required dimensions 'x', 'y', and 'z'. Skipping this variable."
            )
            continue
        data = dataset[var].values
        grid.cell_data[var] = data.flatten(order="C")

    # Take the first data var to check for NaN values and extract only cells in grid
    # that are not NaN. (perhaps this should be done for all data vars?)
    nan_mask = np.isnan(grid.cell_data[grid.cell_data.keys()[0]])
    if np.any(nan_mask):
        grid = grid.extract_cells(~nan_mask)

    return grid
